/*
 * inline.c
 *
 * libhtml - HTML->X renderer
 *
 * Copyright (c) 1995-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "html.h"

typedef struct
{
  WWWDisplay wd;
  WWWData wp;
  WWWRender wn;
  bool setup;
  HTMLInlineInfo ii;
  HTMLInlineState is;
  int width, height, border;
  int vspace, hspace;
  char *mapurl;
  bool ismap;
  HTMLEnv env;
  bool delayed;
} IState;

static void ImgInit _ArgProto((WWWData, void *));
static void ImgAdd _ArgProto((void *));
static void ImgEnd _ArgProto((void *));
static void ImgStop _ArgProto((void *, char *));
static void ImgMessage _ArgProto((void *, char *));

static void DestroyInline _ArgProto((HTMLInfo, HTMLBox));
static bool InlineMouseCallback _ArgProto((void *, int, int, int));
static bool InlineMotionCallback _ArgProto((void *, int, int, int));
static bool InlineResizeCallback _ArgProto((void *));

static void InlineContinue _ArgProto((IState *));

static void SetupInline _ArgProto((HTMLInfo, HTMLBox));
static void AddInline _ArgProto((IState *));
static void CleanIState _ArgProto((IState *));

/*
 * InlineContinue
 */
static void
InlineContinue(isp)
IState *isp;
{
  if (isp->delayed)
  {
    isp->delayed = false;
    HTMLContinueLayout(isp->is.li);
  }
  return;
}

/*
 * SetupInline
 */
static void
SetupInline(li, box)
HTMLInfo li;
HTMLBox box;
{
  IState *isp = (IState *)box->closure;

  isp->setup = true;
  if (isp->wd != NULL)
  {
    WWWSetPosition(isp->wd,
		   box->x + isp->hspace, box->y + isp->vspace,
		   false);
  }

  return;
}

/*
 * CleanIState
 */
static void
CleanIState(isp)
IState *isp;
{
  if (isp->wp != NULL)
  {
    WWWCancelData(isp->wp, true);
    isp->wp = NULL;
  }
  if (isp->wd != NULL)
  {
    WWWDestroyDisplay(isp->wd);
    isp->wd = NULL;
  }
  if (isp->wn != NULL)
  {
    WWWDestroyRender(isp->wn);
    isp->wn = NULL;
  }
  InlineContinue(isp);

  return;
}

/*
 * DestroyInline
 */
static void
DestroyInline(li, box)
HTMLInfo li;
HTMLBox box;
{
  IState *isp = (IState *)box->closure;
  CleanIState(isp);
  return;
}

/*
 * AddInline
 */
static void
AddInline(isp)
IState *isp;
{
  HTMLBox box;
  HTMLAttribID aid;

  isp->is.box = box = HTMLCreateBox(isp->is.li, isp->env);
  
  if ((isp->border = MLAttributeToInt(isp->ii.p, "border")) < 0)
  {
    isp->border = 0;
  }

  aid = HTMLAttributeToID(isp->ii.p, "align");
  if (aid == ATTRIB_MIDDLE) box->baseline = isp->height / 2;
  else if (aid == ATTRIB_TOP) box->baseline = 0;
  else if (aid == ATTRIB_LEFT) HTMLSetB(box, BOX_FLOAT_LEFT);
  else if (aid == ATTRIB_RIGHT) HTMLSetB(box, BOX_FLOAT_RIGHT);
  else box->baseline = isp->height;

  box->setup = SetupInline;
  box->destroy = DestroyInline;
  box->width = isp->width + isp->hspace * 2;
  box->height = isp->height + isp->vspace * 2;
  box->closure = isp;
  
  HTMLEnvAddBox(isp->is.li, isp->env, box);

  return;
}

/*
 * InlineResizeCallback
 */
static bool
InlineResizeCallback(closure)
void *closure;
{
  IState *isp = (IState *)closure;

  if (isp->is.box == NULL)
  {
    WWWGetDimensions(isp->wd, &(isp->width), &(isp->height), false);
    AddInline(isp);
    InlineContinue(isp);
  }

  return(false);
}

/*
 * MouseCallback
 */
static bool
InlineMouseCallback(closure, x, y, button)
void *closure;
int x, y;
int button;
{
  IState *isp = (IState *)closure;
  HTMLAnchor a;
  GList list;
  char *url;

  myassert(isp->is.box != NULL, "MouseCallback: box == NULL");

  if (isp->mapurl != NULL)
  {
    if ((url = HTMLFindMapURL(isp->is.li, isp->mapurl, x, y)) != NULL)
    {
      HTMLLoadURL(isp->is.li, NULL, url, button);
      return(true);
    }
  }

  list = isp->is.li->alist;
  for (a = (HTMLAnchor)GListGetHead(list); a != NULL;
       a = (HTMLAnchor)GListGetNext(list))
  {
    if (a->box == isp->is.box)
    {
      HTMLLoadAnchor(isp->is.li, a, x, y, button, isp->ismap);
      return(true);
    }
  }

  return(true);
}

static bool
InlineMotionCallback(closure, x, y, button)
void *closure;
int x, y;
int button;
{
  IState *isp = (IState *)closure;
  HTMLAnchor a;
  GList list;
  char *url;

  myassert(isp->is.box != NULL, "MotionCallback: box == NULL");

  if (isp->mapurl != NULL)
  {
    if ((url = HTMLFindMapURL(isp->is.li, isp->mapurl, x, y)) != NULL)
    {
      HTMLPrintURL(isp->is.li, url);
      return(true);
    }
  }

  isp->is.li->over = NULL;
  list = isp->is.li->alist;
  for (a = (HTMLAnchor)GListGetHead(list); a != NULL;
       a = (HTMLAnchor)GListGetNext(list))
  {
    if (a->box == isp->is.box)
    {
      HTMLPrintAnchor(isp->is.li, a, x, y, isp->ismap);
      isp->is.li->over = a;
      return(true);
    }
  }

  return(true);
}

/*
 * ImgMessage
 */
static void
ImgMessage(closure, message)
void *closure;
char *message;
{
  IState *isp = (IState *)closure;

  if (isp->is.li->wn != NULL) WWWSendRenderMessage(isp->is.li->wn, message);
  return;
}

/*
 * ImgStop
 */
static void
ImgStop(closure, reason)
void *closure;
char *reason;
{
  IState *isp = (IState *)closure;
  WWWCancelData(isp->wp, true);
  InlineContinue(isp);
  return;
}

/*
 * ImgEnd
 */
static void
ImgEnd(closure)
void *closure;
{
  IState *isp = (IState *)closure;
  WWWEndRender(isp->wn);
  return;
}

/*
 * ImgAdd
 */
static void
ImgAdd(closure)
void *closure;
{
  IState *isp = (IState *)closure;
  WWWAddRender(isp->wn);
  return;
}

/*
 * ImgInit
 */
static void
ImgInit(wp, closure)
WWWData wp;
void *closure;
{
  IState *isp = (IState *)closure;
  WWWRenderHooks *rh;
  char *ctype;
  WWWDisplayCallbacks dcb;

  ctype = WWWGetDataInfo(wp, "content-type");
  if ((rh = WWWGetRenderHooks(WWWDataToContext(wp), ctype)) == NULL)
  {
    WWWCancelData(wp, false);
    InlineContinue(isp);
    return;
  }

  isp->wd = WWWCreateDisplay(isp->is.li->wc, isp->is.li->wd);
  isp->wn = WWWCreateRender(isp->is.li->wc, isp->wd, isp->wp, rh);

  memset(&dcb, 0, sizeof(dcb));
  dcb.motion = InlineMotionCallback;
  dcb.mouse = isp->ii.dcb.mouse;
  dcb.resize = InlineResizeCallback;
  WWWSetDisplayCallbacks(isp->wd, &dcb, isp);

  if (isp->width > 0 && isp->height > 0)
  {
    WWWSetDimensions(isp->wd, isp->width, isp->height, false);
  }
  if (isp->setup)
  {
    WWWSetPosition(isp->wd,
		   isp->is.box->x + isp->hspace, isp->is.box->y + isp->vspace,
		   false);
  }

  return;
}

/*
 *
 * Public functions
 *
 */

void
HTMLCreateInline(li, env, url, ii)
HTMLInfo li;
HTMLEnv env;
char *url;
HTMLInlineInfo *ii;
{
  WWWSinkHooks hooks;
  IState *isp;
  WWWRequest *wr;
  char *xurl;

  if ((wr = WWWBuildRequest(li->wc, url, li->burl)) == NULL)
  {
    return;
  }

  /*
   * Bad kludge to make sure an inline is only reloaded once.
   */
  if (li->reload)
  {
    for (xurl = (char *)GListGetHead(li->loads); xurl != NULL;
	 xurl = (char *)GListGetNext(li->loads))
    {
      if (strlen(wr->url) == strlen(xurl) && strcmp(wr->url, xurl) == 0)
      {
	break;
      }
    }
    if (xurl == NULL)
    {
      WWWReloadRequest(wr, true);
      GListAddHead(li->loads, MPStrDup(li->mp, wr->url));
    }
  }

  /* Select only image content that we can deal with directly */
  if (WWWAddRegexContent(li->wc, wr, "image/*") <= 0)
  {
    WWWDestroyRequest(wr);
    return;
  }

  isp = (IState *)MPCGet(li->mp, sizeof(IState));
  memcpy(&(isp->ii), ii, sizeof(HTMLInlineInfo));
  isp->is.li = li;
  isp->env = env;

  if ((isp->width = MLAttributeToInt(ii->p, "width")) < 0) isp->width = 0;
  if ((isp->height = MLAttributeToInt(ii->p, "height")) < 0) isp->height = 0;
  if ((isp->vspace = MLAttributeToInt(ii->p, "vspace")) < 0) isp->vspace = 4;
  if ((isp->hspace = MLAttributeToInt(ii->p, "hspace")) < 0) isp->hspace = 4;
  isp->ismap = MLFindAttribute(ii->p, "ismap") != NULL ? true:false;
  isp->mapurl = MLFindAttribute(ii->p, "usemap");
 
  if (isp->width == 0 || isp->height == 0)
  {
    isp->delayed = true;
    HTMLDelayLayout(li);
  }
  else AddInline(isp);

  memset(&hooks, 0, sizeof(hooks));
  hooks.stop = ImgStop;
  hooks.init = ImgInit;
  hooks.add = ImgAdd;
  hooks.end = ImgEnd;
  hooks.message = ImgMessage;

  if ((isp->wp = WWWLoadData(li->wc, wr, &hooks, isp)) == NULL)
  {
    InlineContinue(isp);
  }
  GListAddHead(li->datas, isp->wp);

  return;
}

/*
 * HTMLImg
 */
void
HTMLImg(li, env, p)
HTMLInfo li;
HTMLEnv env;
MLElement p;
{
  char *url;
  HTMLInlineInfo ii;

  if ((url = MLFindAttribute(p, "src")) == NULL) return;

  memset(&ii, 0, sizeof(HTMLInlineInfo));
  ii.p = p;
  ii.dcb.mouse = InlineMouseCallback;

  HTMLCreateInline(li, env, url, &ii);

  return;
}
