/*
 * module.c
 *
 * libhtml - HTML->X renderer
 *
 * Copyright (c) 1995-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "html.h"

static bool InAnchor _ArgProto((int, int, HTMLBox));
static void HTMLAdd _ArgProto((void *));
static void HTMLEnd _ArgProto((void *));
static bool HTMLExpose _ArgProto((void *, int, int,
				  unsigned int, unsigned int));
static bool HTMLMotion _ArgProto((void *, int, int, int));
static bool HTMLMouse _ArgProto((void *, int, int, int));
static byte *HTMLQuery _ArgProto((void *, byte *, size_t *));
static int HTMLSearch _ArgProto((void *, byte *, size_t));
static void HTMLDestroy _ArgProto((void *));
static void HTMLCancel _ArgProto((void *));
static void *HTMLInit _ArgProto((WWWRender, void *));
static void HTMLClassDestroy _ArgProto((void *));
int InitModule_HTML _ArgProto((WWWContext));
static void DeselectTimeOut _ArgProto((WWWTimeOut, void *));

static void
DeselectTimeOut(wt, closure)
WWWTimeOut wt;
void *closure;
{
  HTMLAnchor a;
  HTMLInfo li = (HTMLInfo)closure;

  WWWDestroyTimeOut(li->sto);
  li->sto = NULL;

  for (a = (HTMLAnchor)GListGetHead(li->alist); a != NULL;
       a = (HTMLAnchor)GListGetNext(li->alist))
  {
    if (a->p == li->sa->p)
    {
      HTMLClearB(a->box, BOX_SELECT);
      HTMLRenderBox(li, NULL, a->box);
    }
  }

  return;
}

static void
HTMLAdd(closure)
void *closure;
{
  HTMLInfo li = (HTMLInfo)closure;
  byte *data;
  size_t len;
  GList list;

  if (li->cancel) return;

  WWWGetData(li->wp, &data, &len, &list);

  MLAddData(li->hs, (char *)data, len);

  return;
}

static void
HTMLEnd(closure)
void *closure;
{
  HTMLInfo li = (HTMLInfo)closure;
  byte *data;
  size_t len;
  GList list;

  if (li->cancel) return;

  WWWGetData(li->wp, &data, &len, &list);

  MLEndData(li->hs, (char *)data, len);

  return;
}

static bool
HTMLExpose(closure, x, y, width, height)
void *closure;
int x, y;
unsigned int width, height;
{
  HTMLInfo li = (HTMLInfo)closure;
  Region r;
  XRectangle rec;

  r = XCreateRegion();
  rec.x = (short)x;
  rec.y = (short)y;
  rec.width = (unsigned short)width;
  rec.height = (unsigned short)height;
  XUnionRectWithRegion(&rec, r, r);

  HTMLRenderBox(li, r, li->firstbox);

  XDestroyRegion(r);

  return(true);
}

/*
 * InAnchor
 */
static bool
InAnchor(x, y, box)
int x, y;
HTMLBox box;
{
  if (x > box->x &&
      x < box->x + box->width &&
      y > box->y &&
      y < box->y + box->height)
  {
    return(true);
  }

  return(false);
}

static bool
HTMLMouse(closure, x, y, button)
void *closure;
int x, y;
int button;
{
  HTMLInfo li = (HTMLInfo)closure;
  HTMLAnchor a;
  GList list;

  list = li->alist;
  for (a = (HTMLAnchor)GListGetHead(list); a != NULL;
       a = (HTMLAnchor)GListGetNext(list))
  {
    if (InAnchor(x, y, a->box))
    {
      if (a->p != NULL)
      {
	li->sa = a;
	for (a = (HTMLAnchor)GListGetHead(list); a != NULL;
	     a = (HTMLAnchor)GListGetNext(list))
	{
	  if (a->p == li->sa->p)
	  {
	    HTMLSetB(a->box, BOX_SELECT);
	    HTMLRenderBox(li, NULL, a->box);
	  }
	}
	XFlush(li->dpy);
	HTMLLoadAnchor(li, li->sa, x, y, button, false);

	if (li->sto != NULL) WWWDestroyTimeOut(li->sto);
	li->sto = WWWCreateTimeOut(li->wc, li->selectTimeOut,
				   DeselectTimeOut, li);
      }
      break;
    }
  }

  return(true);
}

static bool
HTMLMotion(closure, x, y, button)
void *closure;
int x, y;
int button;
{
  HTMLInfo li = (HTMLInfo)closure;
  HTMLAnchor a;
  GList list;

  list = li->alist;
  for (a = (HTMLAnchor)GListGetHead(list); a != NULL;
       a = (HTMLAnchor)GListGetNext(list))
  {
    if (InAnchor(x, y, a->box)) break;
  }
  
  if (a == NULL || a->p == NULL) WWWSendRenderMessage(li->wn, NULL);
  else HTMLPrintAnchor(li, a, x, y, false);
  li->over = a;

  return(true);
}

/*
 * HTMLQuery
 */
static byte *
HTMLQuery(closure, key, size)
void *closure;
byte *key;
size_t *size;
{
  HTMLInfo li = (HTMLInfo)closure;

  if (strcmp(key, "title") == 0)
  {
    if (li->title != NULL) *size = strlen(li->title);
    else *size = 0;
    return((byte *)li->title);
  }
  return(NULL);
}

/*
 * HTMLSearch
 */
static int
HTMLSearch(closure, data, len)
void *closure;
byte *data;
size_t len;
{
  return(0);
}

/*
 * HTMLCancel
 */
static void
HTMLCancel(closure)
void *closure;
{
  HTMLInfo li = (HTMLInfo)closure;
  WWWData wp;

  myassert(li != NULL, "HTMLInfo is NULL!");

  for (wp = (WWWData)GListGetHead(li->datas); wp != NULL;
       wp = (WWWData)GListGetNext(li->datas))
  {
    WWWCancelData(wp, false);
  }

  li->cancel = true;

  if (li->wt != NULL)
  {
    WWWRemoveTask(li->wc, li->wt);
    li->wt = NULL;
  }

  if (li->sto != NULL) DeselectTimeOut(li->sto, li);

  return;
}

/*
 * HTMLDestroy
 */
static void
HTMLDestroy(closure)
void *closure;
{
  HTMLInfo li = (HTMLInfo)closure;
  WWWData wp;

  HTMLDestroyBox(li, li->firstbox);

  while ((wp = (WWWData)GListPop(li->datas)) != NULL)
  {
    WWWUnloadData(wp);
  }

  if (li->wt != NULL) WWWRemoveTask(li->wc, li->wt);

  if (li->sto != NULL) WWWDestroyTimeOut(li->sto);

  if (li->hs != NULL) MLDestroy(li->hs);

  MPDestroy(li->mp);

  return;
}

/*
 * HTMLInit
 */
static void *
HTMLInit(wn, closure)
WWWRender wn;
void *closure;
{
  HTMLInfo li;
  HTMLClass lc = (HTMLClass)closure;
  MemPool mp;
  WWWContext wc;
  char *url;
  WWWData wp;
  WWWDisplay wd;
  WWWDisplayCallbacks dcb;

  wp = WWWGetRenderData(wn);
  wd = WWWGetRenderDisplay(wn);

  url = WWWGetDataInfo(wp, "x-url");

  mp = MPCreate();
  li = (HTMLInfo)MPCGet(mp, sizeof(struct HTMLInfoP));
  li->mp = mp;
  wc = li->wc = WWWGetRenderContext(wn);
  li->url = url;
  li->burl = url;
  li->wd = wd;
  li->wp = wp;
  li->wn = wn;
  li->hs = MLInit(HTMLHandler, li);
  li->lc = lc;
  li->envstack = GListCreateX(mp);
  li->endstack = GListCreateX(mp);
  li->alist = GListCreateX(mp);
  li->maps = GListCreateX(mp);
  li->datas = GListCreateX(mp);
  li->loads = GListCreateX(mp);

  if (WWWDataWasReloaded(wp)) li->reload = true;
  else li->reload = false;

  WWWGetIntResource(wc, "html.leftMargin", &(li->lmargin));
  if (li->lmargin <= 0) li->lmargin = 20;

  WWWGetIntResource(wc, "html.rightMargin", &(li->rmargin));
  if (li->rmargin <= 0) li->rmargin = 20;

  WWWGetIntResource(wc, "html.bottomMargin", &(li->bmargin));
  if (li->bmargin <= 0) li->bmargin = 20;

  WWWGetIntResource(wc, "html.topMargin", &(li->tmargin));
  if (li->tmargin <= 0) li->tmargin = 20;

  WWWGetBoolResource(wc, "html.printHTMLErrors", &(li->printHTMLErrors));

  li->textLineSpace = -1;
  WWWGetIntResource(wc, "html.textLineSpace", &(li->textLineSpace));
  if (li->textLineSpace < 0) li->textLineSpace = 3;

  li->tableCellInfinity = 0;
  WWWGetIntResource(wc, "html.tableCellInfinity", &(li->tableCellInfinity));

  WWWGetIntResource(wc, "html.inlineTimeOut", &(li->inlineTimeOut));
  if (li->inlineTimeOut < 5) li->inlineTimeOut = 5000;

  WWWGetIntResource(wc, "html.selectTimeOut", &(li->selectTimeOut));
  if (li->selectTimeOut < 1000) li->selectTimeOut = 1000;

  WWWGetBoolResource(li->wc, "html.flowDebug", &(li->flowDebug));

  WWWGetBoolResource(li->wc, "html.constraintDebug", &(li->constraintDebug));

  WWWGetBoolResource(li->wc, "html.printTags", &(li->printTags));

  /*
   * Graphics setup
   */
  li->w = (Widget)WWWGetDrawWidget(wd, true);
  li->dpy = XtDisplay(li->w);
  li->win = XtWindow(li->w);
  li->gc = DefaultGC(li->dpy, DefaultScreen(li->dpy));

  WWWGetNamedColor(wd, "blue", &(li->anchor_color));
  WWWGetNamedColor(wd, "red", &(li->anchor_select_color));
  WWWGetNamedColor(wd, "black", &(li->fg));

  XSetForeground(li->dpy, li->gc, li->fg);
  
  HTMLSetupFonts(li);

  WWWSetScrollBar(wd, true);
  WWWResetDisplay(wd);
  if (WWWGetDimensions(wd, &(li->maxwidth), &(li->maxheight), true) == -1)
  {
    /* Dimensions of display not set so we'll set them. */
    li->maxwidth = 200;
    li->maxheight = 200;
    WWWSetDimensions(wd, li->maxwidth, li->maxheight, false);
    /* Now grab the internal dimensions again */
    if (WWWGetDimensions(wd, &(li->maxwidth), &(li->maxheight), true) == -1)
    {
      return(NULL);
    }
  }

  memset(&dcb, 0, sizeof(dcb));
  dcb.expose = HTMLExpose;
  dcb.mouse = HTMLMouse;
  dcb.motion = HTMLMotion;
  WWWSetDisplayCallbacks(wd, &dcb, li);

  HTMLStart(li);

  return(li);
}

/*
 * HTMLClassDestroy
 */
static void
HTMLClassDestroy(closure)
void *closure;
{
  HTMLClass lc = (HTMLClass)closure;

  HTMLFreeFonts(lc);
  MPDestroy(lc->mp);

  return;
}

/*
 * InitModule_HTML
 */
int
InitModule_HTML(wc)
WWWContext wc;
{
  WWWRenderHooks rh;
  HTMLClass lc;
  MemPool mp;

  mp = MPCreate();
  lc = (HTMLClass)MPCGet(mp, sizeof(struct HTMLClassP));
  lc->mp = mp;

  memset(&rh, 0, sizeof(rh));
  rh.class_context = lc;
  rh.class_destroy = HTMLClassDestroy;
  rh.content = "text/html";
  rh.init = HTMLInit;
  rh.add = HTMLAdd;
  rh.end = HTMLEnd;
  rh.search = HTMLSearch;
  rh.destroy = HTMLDestroy;
  rh.query = HTMLQuery;
  rh.cancel = HTMLCancel;
  WWWAddRenderHooks(wc, &rh);

  return(0);
}
