/*
 * main.c
 *
 * Copyright (C) 1993-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <signal.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Toggle.h>

#include "TextField.h"
#include "AuthDialog.h"
#include "MyDialog.h"

#include "WWW.h"
#include "resources.h"

static void sigigh_handler _ArgProto(());
static void xtWarningHandler _ArgProto((String));
static void xtErrorHandler _ArgProto((String));
static int xErrorHandler _ArgProto((Display *, XErrorEvent *));
static void MakeHead _ArgProto((char *, char *));
static void DeleteAction(), ReturnAction();
static void AuthUICallback _ArgProto((WWWContext, WWWAuthUI,
				      char *, char *, void *));

AppResources appr;

extern char *fallback_resources[];

static XtActionsRec actionsList[] =
{
  { "ReturnAction", (XtActionProc)ReturnAction },
  { "DeleteAction",  (XtActionProc)DeleteAction  },
};

void
main(argc, argv)
int argc;
char **argv;
{
  char *logfile;
  char *dbfiles;
  MemPool mp;
  static char *chimera_fallbacks = "\n\
cache.directory: /tmp\n\
http.userAgent: Chimera/2.0alpha\n\
";

  memset(&appr, 0, sizeof(appr));

  signal(SIGINT, sigigh_handler);
  signal(SIGQUIT, sigigh_handler);
  signal(SIGHUP, sigigh_handler);
  signal(SIGTERM, sigigh_handler);
  signal(SIGPIPE, SIG_IGN);

  StartReaper();

  /*
   * Initialize the Xt stuff.
   */
  XtToolkitInitialize();

  appr.appcon = XtCreateApplicationContext();
  XSetErrorHandler(xErrorHandler);
  XtAppSetErrorHandler(appr.appcon, xtErrorHandler);
  XtAppSetWarningHandler(appr.appcon, xtWarningHandler);

  XtAppSetFallbackResources(appr.appcon, fallback_resources);
  appr.dpy = XtOpenDisplay(appr.appcon, NULL,
                           NULL, "Chimera",
                           NULL, 0,
                           &argc, argv);
  if (appr.dpy == NULL)
  {
    fprintf (stderr, "Could not open display.\n");
    exit(1);
  }

  XtAppAddActions(appr.appcon, actionsList, XtNumber(actionsList));

  /*
   * Initialize the WWW stuff.
   */
  appr.heads = GListCreate();
  appr.downloads = GListCreate();
  appr.views = GListCreate();

  if ((dbfiles = getenv("CHIMERA_DBFILES")) == NULL)
  {
    dbfiles = "~/.chimera/resources";
  }
  appr.wc = WWWCreateContext(appr.appcon, chimera_fallbacks, dbfiles);

  WWWProvideAuthUI(appr.wc, AuthUICallback, NULL);

  appr.bc = BookmarkCreateContext();

  InitChimeraBuiltins(appr.wc);

  mp = MPCreate();
  if ((logfile = WWWGetFileResource(appr.wc, mp,
				    "chimera.urlLogFile")) != NULL)
  {
    appr.logfp = fopen(logfile, "a");
  }
  MPDestroy(mp);

  appr.plain = WWWGetRenderHooks(appr.wc, "text/plain");

  if (argc > 1) MakeHead(argv[1], NULL);
  else MakeHead(NULL, NULL);

  /*
   * And away we go...
   */
  XtAppMainLoop(appr.appcon);
}

/*
 * DeleteAction
 */
static void
DeleteAction()
{
  CHead *h;

  for (h = (CHead *)GListGetHead(appr.heads); h != NULL; )
  {
    DestroyHead(h);
  }

  /* should never get this far */
  return;
}

/*
 * ReturnAction
 */
static void
ReturnAction(w, xe, params, num_params)
Widget w;
XEvent *xe;
String *params;
Cardinal *num_params;
{
  CHead *c;
  char *url;

  for (c = (CHead *)GListGetHead(appr.heads); c != NULL;
       c = (CHead *)GListGetNext(appr.heads))
  {
    if (c->url == w)
    {
      url = TextFieldGetString(c->url);
      if (url == NULL || url[0] == '\0')
      {
	TextFieldSetString(c->url, url);
	break;
      }
      else
      {
	HeadOpen(c, WWWBuildRequest(appr.wc, url, NULL));
      }
      break;
    }
  }

  return;
}

/*
 * MakeHead
 *
 * Frontend that sorts out which URL to use for a new head and then
 * creates a new head.
 */
static void
MakeHead(first, lastresort)
char *first;
char *lastresort;
{
  CHead *h;
  char *buffer;
  char *url;
  int count;
  MemPool mp;
  WWWRequest *wr = NULL;

  h = CreateHead(appr.wc, "chimera");

  /*
   * If an override URL is supplied then try to use that.
   */
  if (first != NULL)
  {
    if ((wr = WWWBuildRequest(appr.wc, first, NULL)) == NULL)
    {
      fprintf (stderr, "Invalid URL: %s\n", first);
    }
  }

  /*
   * Look inside the cutbuffer for a valid URL.
   */
  if (wr == NULL)
  {
    buffer = XFetchBytes(appr.dpy, &count);
    if (count > 0)
    {
      mp = MPCreate();
      url = (char *)MPCGet(mp, count + 1);
      memcpy(url, buffer, count);
      url[count] = '\0';
      
      XFree(buffer);
      
      if ((wr = WWWBuildRequest(appr.wc, url, NULL)) != NULL)
      {
	XStoreBytes(appr.dpy, "", 0);
      }
      
      MPDestroy(mp);
    }
  }

  /*
   * Look for the caller-supplied last resort.
   */
  if (wr == NULL)
  {
    if (lastresort != NULL)
    {
      if ((wr = WWWBuildRequest(appr.wc, lastresort, NULL)) == NULL)
      {
	fprintf (stderr, "Last resort URL (%s) is invalid.\n", lastresort);
      }
    }
  }

  /*
   * Check the standard environment variable.
   */
  if (wr == NULL)
  {
    if ((url = getenv("WWW_HOME")) != NULL)
    {
      if ((wr = WWWBuildRequest(appr.wc, url, NULL)) == NULL)
      {
	fprintf (stderr, "WWW_HOME (%s) is invalid.\n", url);
      }
    }
  }

  /*
   * Check the chimera resource variable.
   */
  if (wr == NULL)
  {
    if ((url = WWWGetResource(appr.wc, "chimera.homeURL")) != NULL)
    {
      if ((wr = WWWBuildRequest(appr.wc, url, NULL)) == NULL)
      {
	fprintf (stderr, "chimera.homeURL (%s) is invalid.\n", url);
      }
    }
  }

  /*
   * This better work...
   */
  if (wr == NULL)
  {
    wr = WWWBuildRequest(appr.wc, "file:/", NULL);
  }

  myassert(wr != NULL, "ERROR: No valud URLs found for new head.\n");

  HeadOpen(h, wr);

  return;
}

/*
 * OOpen
 */
static void
OOpen(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  char *url;

  if ((url = MyDialogGetValue(GetDialogWidget(h->openpop))) == NULL) return;

  HeadOpen(h, WWWBuildRequest(appr.wc, url, NULL));
  XtPopdown(h->openpop); 

  return;
}

/*
 * DOpen
 */
static void
DOpen(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  XtPopdown(h->openpop); 
  return;
}

/*
 * Open
 */
void
Open(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;

  if (h->openpop == NULL)
  {
    h->openpop = CreateDialog(h->toplevel, "openpop",
                              OOpen, DOpen, OOpen, (XtPointer)h);
  }
  XtPopup(h->openpop, XtGrabNone);

  return;
}

/*
 * Back
 */
void
Back(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  HeadBack(h);
  return;
}

/*
 * Reload
 */
void
Reload(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  HeadReload(h);
  return;
}

/*
 * Cancel
 */
void
Cancel(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  HeadCancel(h);
  return;
}

/*
 * Quit
 */
void
Quit(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  DestroyHead((CHead *)cldata);
  return;
}

/*
 * Source
 */
void
Source(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;

  if (appr.plain == NULL) XtVaSetValues(w, XtNstate, False, NULL);
  else
  {
    if (cbdata) h->useplain = true;
    else h->useplain = false;
    HeadRedraw(h);
  }

  return;
}

/*
 * Save
 */
void
Save(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  char *url;
  WWWRequest *wr;

  if ((url = HeadGetURL(h)) == NULL) return;

  if ((wr = WWWBuildRequest(appr.wc, url, NULL)) == NULL)
  {
    TextFieldSetString(h->message, "Invalid URL.");
    return;
  }
  DownloadOpen(wr);

  return;
}

/*
 * Home
 */
void
Home(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  HeadHome(h);
  return;
}

/*
 * Help
 */
void
Help(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  char *url;

  if ((url = WWWGetResource(appr.wc, "chimera.helpURL")) != NULL)
  {
    HeadOpen(h, WWWBuildRequest(appr.wc, url, NULL));
  }

  return;
}

/*
 * sigiqh_handler
 */
static void
sigigh_handler()
{
  return;
}

/*
 * Dup
 */
void
Dup(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  char *url;
  size_t size;

  /* grab URL from to-be-cloned window */
  if ((url = WWWGetResource(appr.wc, "chimera.cloneHome")) == NULL)
  {
    url = (char *)WWWQueryRender(h->wn, "url", &size);
  }

  MakeHead(NULL, url);

  return;
}

/*
 * AddMark
 */
void
AddMark(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  char *title;
  char *url;
  size_t size;

  if (appr.bc == NULL) return;

  if ((url = (char *)WWWQueryRender(h->wn, "url", &size)) == NULL) return;
  if ((title = (char *)WWWQueryRender(h->wn, "title", &size)) == NULL)
  {
    title = url;
  }

  BookmarkAdd(appr.bc, title, url);

  return;
}

/*
 * ViewMark
 */
void
ViewMark(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;
  MemPool mp;
  char *filename;
  char *url;
  const char *fformat = "file:%s";

  if ((filename = WWWGetResource(appr.wc, "bookmark.filename")) == NULL)
  {
    return;
  }

  mp = MPCreate();
  filename = FixPath(mp, filename);
  url = (char *)MPGet(mp, strlen(fformat) + strlen(filename) + 1);
  sprintf (url, fformat, filename);
  HeadOpen(h, WWWBuildRequest(appr.wc, url, NULL));
  MPDestroy(mp);

  return;
}

/*
 * Bookmark
 */
void
Bookmark(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  CHead *h = (CHead *)cldata;

  appr.bmhead = h;
  if (appr.bc != NULL) BookmarkShow(appr.bc);

  return;
}

/*
 * Search
 */
void
Search(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  return;
}

/*
 * xErrorHandler
 */
static int
xErrorHandler(dpy, xe)
Display *dpy;
XErrorEvent *xe;
{
  fprintf (stderr, "X error\n");
  fflush(stderr);
  return(0);
}

/*
 * xtErrorHandler
 */
static void
xtErrorHandler(msg)
String msg;
{
  fprintf (stderr, "%s\n", msg);
  fflush(stderr);
  return;
}

/*
 * xtWarningHandler
 */
static void
xtWarningHandler(msg)
String msg;
{
  fprintf (stderr, "%s\n", msg);
  fflush(stderr);
  return;
}

/*
 * AuthOKCallback
 */
static void
AuthOKCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWAuthUI wu = (WWWAuthUI)cldata;
  char *username;
  char *password;

  username = AuthDialogGetUsername(XtParent(w));
  password = AuthDialogGetPassword(XtParent(w));

  WWWReturnAuth(wu, username, password);
  XtDestroyWidget(XtParent(XtParent(w)));

  return;
}

/*
 * AuthActivateCallback
 */
static void
AuthActivateCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWAuthUI wu = (WWWAuthUI)cldata;
  char *username;
  char *password;

  username = AuthDialogGetUsername(w);
  password = AuthDialogGetPassword(w);

  WWWReturnAuth(wu, username, password);
  XtDestroyWidget(XtParent(w));

  return;
}

/*
 * AuthDismissCallback
 */
static void
AuthDismissCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  WWWAuthUI wu = (WWWAuthUI)cldata;

  WWWReturnAuth(wu, NULL, NULL);
  XtDestroyWidget(XtParent(XtParent(w)));

  return;
}

/*
 * AuthUICallback
 */
static void
AuthUICallback(wc, wu, message, username, closure)
WWWContext wc;
WWWAuthUI wu;
char *message;
char *username;
void *closure;
{
  Widget w, dw;
  Window rw, cw;
  int rx, ry, wx, wy;
  unsigned int mask;

  XQueryPointer(appr.dpy, DefaultRootWindow(appr.dpy),
                &rw, &cw,
		&rx, &ry,
		&wx, &wy,
		&mask);
  w = XtVaAppCreateShell("authorization", "Authorization",
			 topLevelShellWidgetClass,
			 appr.dpy,
			 XtNx, rx - 2,
			 XtNy, ry - 2,
			 NULL);
  dw = XtVaCreateManagedWidget("dialog",
			       authdialogWidgetClass, w,
			       XtNlabel, message,
			       "username", username,
			       NULL);
  MyDialogAddButton(dw, "ok", AuthOKCallback, wu);
  MyDialogAddButton(dw, "dismiss", AuthDismissCallback, wu);
  XtAddCallback(dw, XtNcallback, AuthActivateCallback, wu);
  
  XtRealizeWidget(w);

  return;
}

/*
 * CheckForDone
 */
void
CheckForDone()
{
  if (GListEmpty(appr.heads) && GListEmpty(appr.downloads) &&
      GListEmpty(appr.views))
  {
    if (appr.bc != NULL) BookmarkDestroyContext(appr.bc);

    WWWDestroyContext(appr.wc);

    if (appr.logfp != NULL) fclose(appr.logfp);

    GListDestroy(appr.heads);
    GListDestroy(appr.views);
    GListDestroy(appr.downloads);

    MPPrintStatus();
    GListPrintStatus();

    exit(0);
  }

  return;
}
