/*
 * stack.c
 *
 * Copyright (c) 1995-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "TextField.h"

#include "WWW.h"
#include "resources.h"

extern AppResources appr;

typedef struct
{
  MemPool mp;
  char *url;
  int x, y;
} URLInfo;

static bool StackPop _ArgProto((CHead *));
static void StackCreate _ArgProto((WWWRequest *, StackAction, CHead *));
static void StackInit _ArgProto((WWWData, void *));
static void StackAdd _ArgProto((void *));
static void StackEnd _ArgProto((void *));
static void StackMessage _ArgProto((void *, char *));
static void StackStop _ArgProto((void *, char *));

/*
 * StackPop
 */
static bool
StackPop(h)
CHead *h;
{
  void *muck;

  if ((muck = GListPop(h->urls)) != NULL)
  {
    GListAddHead(h->ourls, muck);
    return(true);
  }

  return(false);
}

/*
 * StackStop
 */
void
StackStop(closure, reason)
void *closure;
char *reason;
{
  return;
}

/*
 * StackMessage
 */
void
StackMessage(closure, message)
void *closure;
char *message;
{
  CHead *h = (CHead *)closure;

  if (message == NULL) message = "";
  TextFieldSetString(h->message, message);

  return;
}

/*
 * StackEnd
 */
void
StackEnd(closure)
void *closure;
{
  CHead *h = (CHead *)closure;
  URLInfo *u;

  WWWEndRender(h->wn);
  if ((u = (URLInfo *)GListGetHead(h->urls)) != NULL)
  {
    WWWSetPosition(h->wd, u->x, u->y, true);
  }

  return;
}

/*
 * StackAdd
 */
void
StackAdd(closure)
void *closure;
{
  CHead *h = (CHead *)closure;

  WWWAddRender(h->wn);
  return;
}

/*
 * HeadOpen
 */
void
HeadOpen(h, wr)
CHead *h;
WWWRequest *wr;
{
  StackCreate(wr, StackNew, h);
  return;
}

/*
 * HeadHome
 */
void
HeadHome(h)
CHead *h;
{
  URLInfo *u;
  
  if (GListGetTail(h->urls) == GListGetHead(h->urls)) return;
  if ((u = (URLInfo *)GListGetTail(h->urls)) == NULL) return;
  StackCreate(WWWBuildRequest(appr.wc, u->url, NULL), StackHome, h);

  return;
}

/*
 * HeadBack
 */
void
HeadBack(h)
CHead *h;
{
  URLInfo *u;

  if (GListGetHead(h->urls) == NULL) return;
  if ((u = (URLInfo *)GListGetNext(h->urls)) == NULL) return;
  StackCreate(WWWBuildRequest(appr.wc, u->url, NULL), StackBack, h);

  return;
}

/*
 * HeadReload
 */
void
HeadReload(h)
CHead *h;
{
  URLInfo *u;
  WWWRequest *wr;

  if ((u = GListGetHead(h->urls)) == NULL) return;
  wr = WWWBuildRequest(appr.wc, u->url, NULL);
  WWWReloadRequest(wr, true);
  StackCreate(wr, StackReload, h);

  return;
}

/*
 * HeadRedraw
 */
void
HeadRedraw(h)
CHead *h;
{
  URLInfo *u;

  if ((u = GListGetHead(h->urls)) == NULL) return;
  StackCreate(WWWBuildRequest(appr.wc, u->url, NULL), StackReplace, h);

  return;
}

/*
 * HeadCancel
 */
void
HeadCancel(h)
CHead *h;
{
  if (h->wp != NULL) WWWCancelData(h->wp, false);
  if (h->pending != NULL) WWWCancelData(h->pending, false);
  if (h->wn != NULL) WWWCancelRender(h->wn);

  return;
}

/*
 * StackInit
 */
static void
StackInit(wp, closure)
WWWData wp;
void *closure;
{
  CHead *h = (CHead *)closure;
  int x, y;
  URLInfo *u;
  char *url;
  char *ctype;
  WWWRenderHooks *rh;
  WWWRequest *wr;
  WWWRenderCallbacks rcb;
  WWWDisplayCallbacks dcb;

  myassert(wp == h->pending, "Data init is not the same as data pending.");

  url = WWWGetDataInfo(wp, "x-url");
  ctype = WWWGetDataInfo(wp, "content-type");
  
  if ((rh = WWWGetRenderHooks(appr.wc, ctype)) == NULL)
  {
    if ((wr = WWWBuildRequest(appr.wc, url, NULL)) == NULL)
    {
      TextFieldSetString(h->message, "Invalid URL.");
    }
    else ViewOpen(wr);
    WWWUnloadData(wp);
    h->pending = NULL;
    return;
  }
  
  if (h->wn != NULL)
  {
    WWWDestroyRender(h->wn);
    h->wn = NULL;
    WWWClearDisplayCallbacks(h->wd);
  }
  
  if (h->wp != NULL) WWWUnloadData(h->wp);
  h->wp = h->pending;
  h->pending = NULL;

  if (h->useplain && appr.plain != NULL)
  {
    if ((h->wn = WWWCreateRender(appr.wc, h->wd,
				 h->wp, appr.plain)) == NULL)
    {
      return;
    }
  }
  else
  {
    if ((h->wn = WWWCreateRender(appr.wc, h->wd, h->wp, rh)) == NULL) return;
  }

  memset(&rcb, 0, sizeof(rcb));
  rcb.action = ActionCallback;
  rcb.message = MessageCallback;
  WWWSetRenderCallbacks(h->wn, &rcb, h);

  memset(&dcb, 0, sizeof(dcb));
  dcb.resize = ResizeCallback;
  dcb.redraw = RedrawCallback;
  WWWSetDisplayCallbacks(h->wd, &dcb, h);

  if (h->action == StackBack)
  {
    StackPop(h);
    u = (URLInfo *)GListGetHead(h->urls);
    x = u->x;
    y = u->y;
    StackPop(h);
  }
  else if (h->action == StackReload || h->action == StackReplace)
  {
    u = (URLInfo *)GListGetHead(h->urls);
    x = u->x;
    y = u->y;
    StackPop(h);
  }
  else if (h->action == StackHome)
  {
    u = (URLInfo *)GListGetTail(h->urls);
    x = u->x;
    y = u->y;
    while (StackPop(h))
	;
    MPDestroy(h->tmp);
    h->tmp = MPCreate();
  }
  else
  {
    if ((u = (URLInfo *)GListGetHead(h->urls)) != NULL)
    {
      WWWGetPosition(h->wd, &(u->x), &(u->y), true);
    }
    x = 0;
    y = 0;
  }

  if ((u = (URLInfo *)GListPop(h->ourls)) == NULL)
  {
    u = (URLInfo *)MPCGet(h->mp, sizeof(URLInfo));
  }
  u->x = x;
  u->y = y;
  u->url = MPStrDup(h->tmp, url);

  TextFieldSetString(h->url, url);

  GListAddHead(h->urls, u);

  return;
}

/*
 * HeadGetURL
 */
char *
HeadGetURL(h)
CHead *h;
{
  URLInfo *u;

  if ((u = (URLInfo *)GListGetHead(h->urls)) == NULL) return(NULL);
  return(u->url);
}

/*
 * StackCreate 
 */
static void
StackCreate(wr, action, h)
WWWRequest *wr;
StackAction action;
CHead *h;
{
  WWWSinkHooks hooks;

  if (wr == NULL)
  {
    TextFieldSetString(h->message, "Invalid URL");
    return;
  }

  HeadCancel(h);
  if (h->pending != NULL)
  {
    WWWUnloadData(h->pending);
    h->pending = NULL;
  }

  h->action = action;

  memset(&hooks, 0, sizeof(hooks));
  hooks.stop = StackStop;
  hooks.init = StackInit;
  hooks.add = StackAdd;
  hooks.end = StackEnd;
  hooks.message = StackMessage;

  /* First, pick all of the contents we can deal with directly */
  WWWAddRegexContent(appr.wc, wr, "*/*");
  /* Add whatever we can get onto the end */
  WWWAddContent(appr.wc, wr, "*/*");

  if (wr->url != NULL && appr.logfp != NULL)
  {
    fprintf (appr.logfp, "%s\n", wr->url);
  }

  h->pending = WWWLoadData(appr.wc, wr, &hooks, h);

  return;
}
