/*
 * view.c
 *
 * Copyright (c) 1995-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "TextField.h"
#include "MyDialog.h"

#include "WWW.h"
#include "resources.h"

extern AppResources appr;

typedef struct 
{
  MemPool mp;
  WWWData wp;
  char *command;
  WWWTask wt;
  WWWContext wc;
} ViewInfo;

static char *ViewFind _ArgProto((MemPool, char *, char *));
static void ViewDestroy _ArgProto((void *));
static void DoViewDestroy _ArgProto((ViewInfo *));
static void ViewInit _ArgProto((WWWData, void *));
static void ViewAdd _ArgProto((void *));
static void ViewEnd _ArgProto((void *));
static void ViewStop _ArgProto((void *, char *));

/*
 * ViewFind
 *
 * Return a command string for a type.
 */
static char *
ViewFind(mp, filelist, content)
MemPool mp;
char *filelist;
char *content;
{
  FILE *fp;
  char *f;
  char buffer[BUFSIZ];
  char ctype[BUFSIZ];
  char command[BUFSIZ];
  char junk[BUFSIZ];
  char *filename;
  MemPool tmp;

  tmp = MPCreate();

  f = filelist;
  while ((filename = mystrtok(f, ':', &f)) != NULL)
  {
    filename = FixPath(mp, filename);
    if (filename == NULL) continue;
   
    fp = fopen(filename, "r");
    if (fp == NULL) continue;
    
    while (fgets(buffer, sizeof(buffer), fp))
    {
      if (buffer[0] == '#' || buffer[0] == '\n') continue;
      
      if (sscanf(buffer, "%[^;]%[;]%[^;\n]", ctype, junk, command) == 3)
      {
	if (WWWMatchContent(tmp, ctype, content))
        {
	  fclose(fp);
          MPDestroy(tmp);
	  return(MPStrDup(mp, command));
	}
      }
    }

    fclose(fp);
  }

  MPDestroy(tmp);

  return(NULL);
}

static void
DoViewDestroy(vi)
ViewInfo *vi;
{
  if (vi->wt != NULL) return;

  vi->wt = WWWScheduleTask(vi->wc, ViewDestroy, vi);

  return;
}

void
ViewEnd(closure)
void *closure;
{
  ViewInfo *vi = (ViewInfo *)closure;
  char *filename;
  FILE *fp;
  char *final;
  char *path;
  byte *data;
  size_t len;
  GList mimelist;
  size_t finallen;

  WWWGetData(vi->wp, &data, &len, &mimelist);
  if (data == NULL || len == 0)
  {
    DoViewDestroy(vi);
    return;
  }

  filename = mytmpnam(vi->mp);
  if ((fp = fopen(filename, "w")) != NULL)
  {
    fwrite(data, len, 1, fp);
    fclose(fp);

    if ((path = WWWGetResource(appr.wc, "view.path")) == NULL) path = "";
    
    finallen = strlen(" ; rm &") + strlen(path) + strlen(filename) * 2 +
	strlen(vi->command) + strlen("PATH=") + 1;
    final = (char *)MPGet(vi->mp, finallen);
    strcpy(final, "PATH=");
    strcat(final, path);
    strcat(final, " ");
    sprintf (final + strlen(final), vi->command, filename);
    strcat(final, "; rm ");
    strcat(final, filename);
    strcat(final, "&");
    system(final);
  }

  DoViewDestroy(vi);

  return;
}

/*
 * ViewDestroy
 */
static void
ViewDestroy(closure)
void *closure;
{
  ViewInfo *vi = (ViewInfo *)closure;

  WWWUnloadData(vi->wp);
  MPDestroy(vi->mp);

  CheckForDone();

  return;
}

/*
 * ViewAdd
 */
void
ViewAdd(closure)
void *closure;
{
  return;
}

/*
 * ViewStop
 */
void
ViewStop(closure, reason)
void *closure;
char *reason;
{
  return;
}

/*
 * ViewInit
 */
static void
ViewInit(wp, closure)
WWWData wp;
void *closure;
{
  char *url;
  char *caplist;
  ViewInfo *vi = (ViewInfo *)closure;
  WWWRequest *wr;

  if ((caplist = WWWGetResource(appr.wc, "view.capFiles")) == NULL)
  {
    caplist = "~/.mailcap";
  }

  if ((vi->command = ViewFind(vi->mp, caplist,
			      WWWGetDataInfo(wp, "content-type"))) == NULL)
  {
    DoViewDestroy(vi);
    url = WWWGetDataInfo(wp, "x-url");

    if ((wr = WWWBuildRequest(appr.wc, url, NULL)) == NULL) return;
    DownloadOpen(wr);

    return;
  }

  return;
}

/*
 * ViewOpen
 */
void
ViewOpen(wr)
WWWRequest *wr;
{
  ViewInfo *vi;
  WWWSinkHooks hooks;
  MemPool mp;

  myassert(wr != NULL, "Request cannot be NULL.");

  mp = MPCreate();
  vi = (ViewInfo *)MPCGet(mp, sizeof(ViewInfo));
  vi->mp = mp;
  vi->wc = appr.wc;

  memset(&hooks, 0, sizeof(hooks));
  hooks.stop = ViewStop;
  hooks.init = ViewInit;
  hooks.add = ViewAdd;
  hooks.end = ViewEnd;

  if ((vi->wp = WWWLoadData(appr.wc, wr, &hooks, vi)) != NULL)
  {
    GListAddHead(appr.views, vi);
  }
  else DoViewDestroy(vi);

  return;
}
