/*
 * data.c
 *
 * Copyright (c) 1996-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include "WWWP.h"
#include "mime.h"

typedef enum
{
  DataStateWaiting,
  DataStateEnded,
  DataStateLoading,
  DataStateUnloaded
} DataState;

typedef enum
{
  SchemeStateBlocked,
  SchemeStateWaiting,
  SchemeStateEnded,
  SchemeStateLame,
  SchemeStateLoading
} SchemeState;

struct WWWDataP
{
  MemPool           mp;
  DataState         dstate;
  WWWScheme         ws;
  WWWSinkHooks      dhooks;
  void              *closure;
  WWWTask           wt;
  bool              locked;
  bool              destroy;
  bool              stop_callbacks;
};

struct WWWSchemeP
{
  MemPool           mp;
  WWWContext        wc;
  WWWSchemeHooks    shooks;
  void              *closure;
  GList		    datas;

  /* only used if the scheme is blocked.  do not free. */
  WWWRequest        *wr;

  /* Various status goodies */
  SchemeState       sstate;             /* transfer state */
  bool              cachable;           /* can this be cached? */
  bool              reusable;           /* can this be reused? */
};

/*
 * Private functions
 */
static WWWScheme SchemeInit _ArgProto((WWWContext, WWWRequest *));
static void SchemeDestroy _ArgProto((WWWContext, WWWScheme));
static void SchemeEndTask _ArgProto((void *));
static void SchemeInitTask _ArgProto((void *));
static void SchemeStopTask _ArgProto((void *));
static char *GetSchemeInfo _ArgProto((WWWScheme, char *));
static void StartBlockedScheme _ArgProto((WWWContext));
static int StartScheme _ArgProto((WWWContext, WWWScheme));
static WWWScheme TryMemoryCache _ArgProto((WWWContext, WWWRequest *));

/*
 * SchemeDestroy
 */
static void
SchemeDestroy(wc, ws)
WWWContext wc;
WWWScheme ws;
{
  WWWData c;

  if (wc->printLoadMessages)
  {
    fprintf (stderr, "Scheme Destroy %s\n", ws->wr->url);
  }

  for (c = (WWWData)GListGetHead(ws->datas); c != NULL;
       c = (WWWData)GListGetNext(ws->datas))
  {
    myassert(c->dstate == DataStateUnloaded, "Data not unloaded");

    if (c->wt != NULL) WWWRemoveTask(ws->wc, c->wt);
    MPDestroy(c->mp);    
  }

  if (ws->closure != NULL) (ws->shooks.destroy)(ws->closure);
  GListRemoveItem(wc->sources, ws);
  MPDestroy(ws->wr->mp);
  MPDestroy(ws->mp);

  return;
}

/*
 * SchemeInit
 */
static WWWScheme
SchemeInit(wc, wr)
WWWContext wc;
WWWRequest *wr;
{
  WWWScheme c;
  int count;
  GList list;
  WWWScheme ws;
  MemPool mp;

  mp = MPCreate();
  ws = (WWWScheme)MPCGet(mp, sizeof(struct WWWSchemeP));
  ws->mp = mp;
  ws->wc = wc;
  ws->cachable = false;
  ws->reusable = false;
  ws->sstate = SchemeStateWaiting;
  ws->datas = GListCreateX(mp);
  ws->wr = wr;

  count = 0;
  list = wc->sources;
  for (c = (WWWScheme)GListGetHead(list); c != NULL;
       c = (WWWScheme)GListGetNext(list))
  {
    if (c->sstate == SchemeStateLoading || c->sstate == SchemeStateWaiting)
    {
      count++;
      if (count == wc->maxiocnt)
      {
	ws->sstate = SchemeStateBlocked;
	GListAddHead(wc->sources, ws);
	if (wc->printLoadMessages)
	{
	  fprintf (stderr, "Scheme Block %s\n", wr->url);
	}
	return(ws);
      }
    }
  }

  GListAddHead(wc->sources, ws);

  if (StartScheme(wc, ws) == -1) return(NULL);

  return(ws);
}

/*
 * StartScheme
 */
static int
StartScheme(wc, ws)
WWWContext wc;
WWWScheme ws;
{
  WWWSchemeHooks *hooks;
  WWWRequest *wr = ws->wr;

  /*
   * Look in the cache if the reload flag isn't set.  Copy the cache
   * scheme hooks to be used later.  Set the reusable flag so that the
   * memory can be used for later loads.
   */
  if (!wr->reload && wr->input_data == NULL && wc->cc != NULL)
  {
    hooks = WWWGetCacheHooks(wc->cc);
    if (hooks != NULL)
    {
      ws->closure = (hooks->init)(ws, wr, hooks->class_closure);
      if (ws->closure != NULL)
      {
	ws->reusable = true;
	memcpy(&(ws->shooks), hooks, sizeof(ws->shooks));
      }
    }
  }

  /*
   * Cache could not be attempted or the cache grab failed.  Try the
   * real handler.
   */
  if (ws->closure == NULL)
  {
    ws->closure = (wr->hooks.init)(ws, wr, wr->hooks.class_closure);
    memcpy(&(ws->shooks), &(wr->hooks), sizeof(ws->shooks));    
  }

  /*
   * Couldn't start a scheme.  Just give up and return nothing.
   */
  if (ws->closure == NULL)
  {
    GListRemoveItem(wc->sources, ws);
    SchemeDestroy(wc, ws);

    if (wc->printLoadMessages)
    {
      fprintf (stderr, "Scheme StartFailed %s\n", wr->url);
    }

    return(-1);
  }

  ws->sstate = SchemeStateWaiting;

  if (wc->printLoadMessages)
  {
    fprintf (stderr, "Scheme StartSuccess %s\n", wr->url);
  }

  return(0);
}

/*
 * StartBlockedScheme
 */
static void
StartBlockedScheme(wc)
WWWContext wc;
{
  WWWScheme c;

  for (c = (WWWScheme)GListGetHead(wc->sources); c != NULL; )
  {
    if (c->sstate == SchemeStateBlocked)
    {
      if (wc->printLoadMessages)
      {
	fprintf (stderr, "Scheme StartBlocked %s\n", c->wr->url);
      }
      if (StartScheme(wc, c) == 0) break;
      else c = (WWWScheme)GListGetHead(wc->sources);
    }
    else c = (WWWScheme)GListGetNext(wc->sources);
  }

  return;
}

/*
 * SchemeStopTask
 */
static void
SchemeStopTask(closure)
void *closure;
{
  WWWData wp = (WWWData)closure;

  wp->wt = NULL;
  if (wp->stop_callbacks) return;

  wp->locked = true;
  
  (wp->dhooks.stop)(wp->closure, "");
  wp->wt = NULL;
  
  wp->locked = false;

  if (wp->destroy) WWWUnloadData(wp);

  return;
}

/*
 * SchemeInitTask
 */
static void
SchemeInitTask(closure)
void *closure;
{
  WWWData wp = (WWWData)closure;

  wp->wt = NULL;
  if (wp->stop_callbacks) return;

  wp->locked = true;
  
  if (WWWMatchRequestContent(wp->ws->wr,
			     GetSchemeInfo(wp->ws, "content-type")))
  {
    wp->dstate = DataStateLoading;
    (wp->dhooks.init)(wp, wp->closure);
    
    if (wp->ws->sstate == SchemeStateEnded)
    {
      wp->wt = WWWScheduleTask(wp->ws->wc, SchemeEndTask, wp);
    }
  }
  else wp->wt = WWWScheduleTask(wp->ws->wc, SchemeStopTask, wp);
  
  wp->locked = false;

  if (wp->destroy) WWWUnloadData(wp);

  return;
}

/*
 * SchemeEndTask
 */
static void
SchemeEndTask(closure)
void *closure;
{
  WWWData wp = (WWWData)closure;

  wp->wt = NULL;
  if (wp->stop_callbacks) return;

  wp->locked = true;
  
  if (wp->dstate == DataStateLoading)
  {
    wp->dstate = DataStateEnded;
    (wp->dhooks.end)(wp->closure);
  }
  
  wp->locked = false;

  if (wp->destroy) WWWUnloadData(wp);

  return;
}

/*
 * TryMemoryCache
 *
 * Rummage around in the list of current downloads to see if what we need
 * is already there.
 */
static WWWScheme
TryMemoryCache(wc, wr)
WWWContext wc;
WWWRequest *wr;
{
  WWWScheme ws;
  char *content;

  /*
   * Input data can make the output different everytime don't even
   * try to find something that has been cached.
   */
  if (wr->input_data != NULL) return(NULL);

  for (ws = (WWWScheme)GListGetHead(wc->sources); ws != NULL;
       ws = (WWWScheme)GListGetNext(wc->sources))
  {
    /*
     * Obviously the URL has to match.
     */
    if (WWWCompareRequestURL(ws->wr, wr))
    {
      /*
       * Check to make sure that it hasn't been cancelled and that its
       * reusable.
       */
      if (ws->sstate != SchemeStateLame && (ws->cachable || ws->reusable))
      {
	/*
	 * Try to get the content from the download.  If the content-type is
	 * unknown then see if the accept list is the same which is
	 * acceptable (zing!).  If the content-type is known then see if it
	 * is in the accept list for the request.
	 */
	if ((content = GetSchemeInfo(ws, "content-type")) == NULL)
	{
	  if (WWWCompareRequestAccept(ws->wr, wr)) return(ws);
	  if (wc->printLoadMessages)
	  {
	    fprintf (stderr, "Data AcceptListDifferent %s\n", wr->url);
	  }
	}
	else if (WWWMatchRequestContent(ws->wr, content)) return(ws);
	else if (wc->printLoadMessages)
	{
	  fprintf (stderr, "Data ContentDifferent %s\n", wr->url);
	}
      }
      else if (wc->printLoadMessages)
      {
	fprintf (stderr, "Data StateBad %s\n", wr->url);
      }
    }
  }
  
  return(NULL);
}

/*
 * WWWLoadData
 *
 * Load some data.  'wr' is eaten by this function.
 */
WWWData
WWWLoadData(wc, wr, hooks, closure)
WWWContext wc;
WWWRequest *wr;
WWWSinkHooks *hooks;
void *closure;
{
  WWWScheme ws = NULL;
  WWWData wp;
  MemPool mp;

  myassert(hooks->init != NULL, "Init function is NULL.");
  myassert(hooks->add != NULL, "Add function is NULL.");
  myassert(hooks->end != NULL, "End function is NULL.");
  myassert(hooks->stop != NULL, "Stop function is NULL.");

  if ((ws = TryMemoryCache(wc, wr)) == NULL)
  {
    if ((ws = SchemeInit(wc, wr)) == NULL)
    {
      WWWDestroyRequest(wr);
      return(NULL);
    }
  }
  else
  {
    /*
     * This can be thrown out since 'ws' contains an identical copy.
     */
    WWWDestroyRequest(wr);
    if (wc->printLoadMessages)
    {
      fprintf (stderr, "Data MemCached %s\n", wr->url);
    }
  }

  mp = MPCreate();
  wp = (WWWData)MPCGet(mp, sizeof(struct WWWDataP));
  wp->mp = mp;
  memcpy(&(wp->dhooks), hooks, sizeof(WWWSinkHooks));
  wp->closure = closure;
  wp->dstate = DataStateWaiting;
  wp->ws = ws;

  GListAddTail(ws->datas, wp);
  
  if (ws->sstate == SchemeStateLoading || ws->sstate == SchemeStateEnded)
  {
    wp->wt = WWWScheduleTask(ws->wc, SchemeInitTask, wp);
  }

  return(wp);
}

/*
 * WWWUnloadData
 */
void
WWWUnloadData(wp)
WWWData wp;
{
  WWWScheme ws = wp->ws;
  WWWData c;

  if (wp->locked)
  {
    wp->destroy = true;
    wp->stop_callbacks = true;
    return;
  }

  if (ws->wc->printLoadMessages)
  {
    fprintf (stderr, "Data Unload %s\n", ws->wr->url);
  }

  wp->dstate = DataStateUnloaded;
  for (c = (WWWData)GListGetHead(ws->datas); c != NULL;
       c = (WWWData)GListGetNext(ws->datas))
  {
    if (c->wt != NULL)
    {
      WWWRemoveTask(ws->wc, c->wt);
      c->wt = NULL;
    }
    if (c->dstate != DataStateUnloaded) break;
  }
  if (c == NULL) SchemeDestroy(ws->wc, ws);

  return;
}

/*
 * WWWDataWasReloaded
 */
bool
WWWDataWasReloaded(wp)
WWWData wp;
{
  return(wp->ws->wr->reload);
}

/*
 * WWWStopScheme
 */
void
WWWStopScheme(ws, reason)
WWWScheme ws;
char *reason;
{
  WWWData c;
  GList list;

  ws->sstate = SchemeStateLame;

  list = ws->datas;
  for (c = (WWWData)GListGetHead(list); c != NULL;
       c = (WWWData)GListGetNext(list))
  {
    if (!c->stop_callbacks) (c->dhooks.stop)(c->closure, reason);
  }

  StartBlockedScheme(ws->wc);

  return;
}

/*
 * WWWEndScheme
 */
void
WWWEndScheme(ws)
WWWScheme ws;
{
  GList list;
  WWWData c;

  myassert(ws->sstate == SchemeStateLoading, "End source in wrong state");

  ws->sstate = SchemeStateEnded;

  list = ws->datas;
  for (c = (WWWData)GListGetHead(list); c != NULL;
       c = (WWWData)GListGetNext(list))
  {
    if (c->dstate == DataStateLoading && !c->stop_callbacks)
    {
      c->dstate = DataStateEnded;
      (c->dhooks.end)(c->closure);
    }
  }

  if (ws->wc->cc && ws->cachable) WWWWriteCache(ws->wc->cc, ws);

  StartBlockedScheme(ws->wc);

  return;
}

/*
 * WWWAddScheme
 */
void
WWWAddScheme(ws)
WWWScheme ws;
{
  GList list;
  WWWData c;

  myassert(ws->sstate == SchemeStateLoading, "Add source in wrong state");

  list = ws->datas;
  for (c = (WWWData)GListGetHead(list); c != NULL;
       c = (WWWData)GListGetNext(list))
  {
    if (c->dstate == DataStateLoading && !c->stop_callbacks)
    {
      (c->dhooks.add)(c->closure);
    }
  }

  return;
}

/*
 * WWWInitScheme
 */
void
WWWInitScheme(ws, usecache)
WWWScheme ws;
bool usecache;
{
  GList list;
  WWWData c;
  char *url;

  myassert(ws->sstate == SchemeStateWaiting, "Init source in wrong state.");
  ws->sstate = SchemeStateLoading;
  ws->cachable = usecache;

  url = GetSchemeInfo(ws, "x-url");
  myassert(url != NULL, "URL is NULL");
  myassert(GetSchemeInfo(ws, "content-type") != NULL,
	   "Content-type is NULL");

  list = ws->datas;
  for (c = (WWWData)GListGetHead(list); c != NULL;
       c = (WWWData)GListGetNext(list))
  {
    c->wt = WWWScheduleTask(ws->wc, SchemeInitTask, c);
  }

  return;
}

/*
 * WWWSendSchemeMessage
 */
void
WWWSendSchemeMessage(ws, message)
WWWScheme ws;
char *message;
{
  GList list;
  WWWData c;

  list = ws->datas;
  for (c = (WWWData)GListGetHead(list); c != NULL;
       c = (WWWData)GListGetNext(list))
  {
    if (c->dhooks.message != NULL && !c->stop_callbacks)
    {
      (c->dhooks.message)(c->closure, message);
    }
  }

  return;
}

/*
 * WWWDataToContext
 */
WWWContext
WWWDataToContext(wp)
WWWData wp;
{
  return(wp->ws->wc);
}

/*
 * WWWSchemeToContext
 */
WWWContext
WWWSchemeToContext(ws)
WWWScheme ws;
{
  return(ws->wc);
}

/*
 * WWWGetData
 */
void
WWWGetData(wp, data, len, list)
WWWData wp;
byte **data;
size_t *len;
GList *list;
{
  (wp->ws->shooks.getdata)(wp->ws->closure, data, len, list);
  return;
}

/*
 * WWWGetDataInfo
 */
char *
WWWGetDataInfo(wp, name)
WWWData wp;
char *name;
{
  byte *data;
  size_t len;
  GList list;

  WWWGetData(wp, &data, &len, &list);
  if (list != NULL) return(MIMEGetListField(list, name));
  return(NULL);
}

/*
 * WWWGetSchemeData
 */
void
WWWGetSchemeData(ws, data, len, list)
WWWScheme ws;
byte **data;
size_t *len;
GList *list;
{
  (ws->shooks.getdata)(ws->closure, data, len, list);
  return;
}

/*
 * GetSchemeInfo
 */
static char *
GetSchemeInfo(ws, name)
WWWScheme ws;
char *name;
{
  byte *data;
  size_t len;
  GList list;

  WWWGetSchemeData(ws, &data, &len, &list);
  if (list == NULL) return(NULL);
  return(MIMEGetListField(list, name));
}

/*
 * WWWCancelData
 */
void
WWWCancelData(wp, just_stop_callbacks)
WWWData wp;
bool just_stop_callbacks;
{
  WWWScheme ws = wp->ws;

  wp->stop_callbacks = true;
  if (wp->wt != NULL)
  {
    WWWRemoveTask(ws->wc, wp->wt);
    wp->wt = NULL;
  }

  /*
   * This should probably be more sophisticated.
   */
  if (!just_stop_callbacks)
  {
    ws->sstate = SchemeStateLame;
    (ws->shooks.stop)(ws->closure);
  }

  return;
}
