/*
 * display.c
 *
 * Copyright (c) 1995-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "WWWP.h"

typedef struct
{
  WWWDisplayCallbacks dcb;
  void *closure;
} DisplayCB;

typedef struct
{
  unsigned int width, height;
  bool internal;
} ResizeTaskInfo;

struct WWWDisplayP
{
  Widget      www;                    /* Xt widget */
  MemPool     mp;                     /* memory pool */
  WWWContext  wc;

  /* state flags */
  bool        position_set;
  bool        size_set;
  bool        use_scroll; 

  GList       cblist;
  WWWTask     resize_task;
  ResizeTaskInfo rti;
  bool        in_resize_task;
};

static void ResizeTask _ArgProto((void *));

/*
 * WWWCreateDisplay
 */
WWWDisplay
WWWCreateDisplay(wc, parent)
WWWContext wc;
WWWDisplay parent;
{
  WWWDisplay wd;
  MemPool mp;
  WWWWidget rw;

  myassert(parent != NULL, "NULL parent not allowed");

  mp = MPCreate();
  wd = (WWWDisplay)MPCGet(mp, sizeof(struct WWWDisplayP));
  wd->mp = mp;
  wd->wc = wc;
  wd->cblist = GListCreateX(mp);

  wd->www = XtVaCreateManagedWidget("www_child",
				    wwwWidgetClass,
				    WWWDrawWidget(parent->www),
				    XtNmappedWhenManaged, False,
				    XtNborderWidth, 0,
				    NULL);

  rw = (WWWWidget)wd->www;
  rw->www.wd = wd;

  return(wd);
}

/*
 * WWWDestroyDisplay
 */
int
WWWDestroyDisplay(wd)
WWWDisplay wd;
{
  if (wd->resize_task != NULL) WWWRemoveTask(wd->wc, wd->resize_task);

  /* clean up misc. resources */
  if (wd->www != NULL) XtDestroyWidget(wd->www);

  MPDestroy(wd->mp);

  return(0);
}

/*
 * WWWDisplayToWidget
 */
void *
WWWDisplayToWidget(wd)
WWWDisplay wd;
{
  return(wd->www);
}

/*
 * WWWExposeDisplay
 */
void
WWWExposeDisplay(wd, x, y, width, height)
WWWDisplay wd;
int x, y;
unsigned int width, height;
{
  DisplayCB *t;

  for (t = (DisplayCB *)GListGetHead(wd->cblist); t != NULL;
       t = (DisplayCB *)GListGetNext(wd->cblist))
  {
    if (t->dcb.expose != NULL)
    {
      if (!(t->dcb.expose)(t->closure, x, y, width, height)) return;
    }
  }

  return;
}

/*
 * WWWSelectDisplay
 */
void
WWWSelectDisplay(wd, x, y, button)
WWWDisplay wd;
int x, y;
int button;
{
  DisplayCB *t;

  for (t = (DisplayCB *)GListGetHead(wd->cblist); t != NULL;
       t = (DisplayCB *)GListGetNext(wd->cblist))
  {
    if (t->dcb.mouse != NULL)
    {
      if (!(t->dcb.mouse)(t->closure, x, y, button)) return;
    }
  }

  return;
}

/*
 * WWWMotionDisplay
 */
void
WWWMotionDisplay(wd, x, y, button)
WWWDisplay wd;
int x, y;
int button;
{
  DisplayCB *t;

  for (t = (DisplayCB *)GListGetHead(wd->cblist); t != NULL;
       t = (DisplayCB *)GListGetNext(wd->cblist))
  {
    if (t->dcb.motion != NULL)
    {
      if (!(t->dcb.motion)(t->closure, x, y, button)) return;
    }
  }

  return;
}

/*
 * WWWCreateToplevelDisplay
 */
WWWDisplay
WWWCreateToplevelDisplay(wc, parent)
WWWContext wc;
void *parent;
{
  WWWDisplay wd;
  WWWWidget rw;
  MemPool mp;

  mp = MPCreate();
  wd = (WWWDisplay)MPCGet(mp, sizeof(struct WWWDisplayP));
  wd->mp = mp;
  wd->cblist = GListCreateX(mp);
  wd->wc = wc;

  wd->www = XtVaCreateManagedWidget("www_toplevel",
				    wwwWidgetClass, (Widget)parent,
				    NULL);
  rw = (WWWWidget)wd->www;
  rw->www.wd = wd;
  wd->size_set = true;

  return(wd);
}

void
WWWRedrawDisplay(wd)
WWWDisplay wd;
{
  DisplayCB *t;

  for (t = (DisplayCB *)GListGetHead(wd->cblist); t != NULL;
       t = (DisplayCB *)GListGetNext(wd->cblist))
  {
    if (t->dcb.redraw != NULL)
    {
      if (!(t->dcb.redraw)(t->closure)) return;
    }
  }

  return;
}

/*
 * WWWSetScrollBar
 */
void
WWWSetScrollBar(wd, use_scroll)
WWWDisplay wd;
bool use_scroll;
{
  wd->use_scroll = use_scroll;

  return;
}

/*
 * WWWGetDimensions
 */
int
WWWGetDimensions(wd, width, height, internal)
WWWDisplay wd;
unsigned int *width, *height;
bool internal;
{
  WWWWidget rw = (WWWWidget)wd->www;

  if (!wd->size_set) return(-1);

  if (internal)
  {
    *width = (unsigned int)rw->www.child->core.width;
    *height = (unsigned int)rw->www.child->core.height;
  }
  else
  {
    WWWGetSize(wd->www, wd->use_scroll, width, height);
  }
     
  return(0);
}

int
WWWSetPosition(wd, x, y, internal)
WWWDisplay wd;
int x, y;
bool internal;
{
  if (internal)
  {
    WWWMoveChild((WWWWidget)wd->www, (Position)-x, (Position)-y);
  }
  else
  {
    wd->position_set = true;
    if (wd->www != NULL)
    {
      XtMoveWidget(wd->www, (Position)x, (Position)y);
      XtSetMappedWhenManaged(wd->www, True);
    }
  }

  return(0);
}

int
WWWGetPosition(wd, x, y, internal)
WWWDisplay wd;
int *x, *y;
bool internal;
{
  if (internal)
  {
    *x = ((WWWWidget)(wd->www))->www.child->core.x;
    *y = ((WWWWidget)(wd->www))->www.child->core.y;
  }
  else
  {
    *x = wd->www->core.x;
    *y = wd->www->core.y;
  }

  return(0);
}

static void
ResizeTask(closure)
void *closure;
{
  DisplayCB *t;
  WWWDisplay wd = (WWWDisplay)closure;
  bool internal = wd->rti.internal;
  unsigned int width = wd->rti.width;
  unsigned int height = wd->rti.height;

  wd->in_resize_task = true;

  if (internal)
  {
    WWWSetDrawSize(wd->www, wd->use_scroll, width, height);
  }
  else
  {
    if (wd->www != NULL)
    {
      WWWSetSize(wd->www, wd->use_scroll, width, height);
      WWWSetDrawSize(wd->www, wd->use_scroll, width, height);
    }
    wd->size_set = true;

    for (t = (DisplayCB *)GListGetHead(wd->cblist); t != NULL;
	 t = (DisplayCB *)GListGetNext(wd->cblist))
    {
      if (t->dcb.resize != NULL)
      {
	if (!(t->dcb.resize)(t->closure)) break;
      }
    }
  }

  wd->in_resize_task = false;
  wd->resize_task = NULL;

  return;
}

void
WWWSetDimensions(wd, width, height, internal)
WWWDisplay wd;
unsigned int width, height;
bool internal;
{
  if (wd->in_resize_task) return;
  if (wd->resize_task != NULL) WWWRemoveTask(wd->wc, wd->resize_task);

  wd->rti.width = width;
  wd->rti.height = height;
  wd->rti.internal = internal;
  wd->resize_task = WWWScheduleTask(wd->wc, ResizeTask, wd);

  return;
}

int
WWWGetNamedColor(wd, name, pixel)
WWWDisplay wd;
char *name;
Pixel *pixel;
{
  XColor sxc, exc;
  Display *dpy = XtDisplay(wd->www);

  XAllocNamedColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)),
                   name, &sxc, &exc);

  *pixel = sxc.pixel;

  return(0);
}

void *
WWWGetDrawWidget(wd, internal)
WWWDisplay wd;
bool internal;
{
  if (internal) return(WWWDrawWidget(wd->www));
  return(wd->www);
}

/*
 * Mmmmm, long name.
 */
void
WWWGetOnScreenDimensions(wd, x, y, width, height)
WWWDisplay wd;
int *x, *y;
unsigned int *width, *height;
{
  WWWWidget rw = (WWWWidget)wd->www;
  
  *x = -(int)rw->www.child->core.x;
  *y = -(int)rw->www.child->core.y;
  *width = (unsigned int)rw->www.child->core.width;
  *height = (unsigned int)rw->www.clip->core.height;
    
  return;
}     

/*
 * WWWResetDisplay
 */
void
WWWResetDisplay(wd)
WWWDisplay wd;
{
  WWWWidget rw = (WWWWidget)wd->www;

  XClearWindow(XtDisplay(rw->www.child), XtWindow(rw->www.child));
  WWWMoveChild(rw, 0, 0);
  WWWSetDrawSize(wd->www, wd->use_scroll, 0, 0);

  return;
}

/*
 * WWWSetDisplayCallbacks
 */
void
WWWSetDisplayCallbacks(wd, dcb, dc)
WWWDisplay wd;
WWWDisplayCallbacks *dcb;
void *dc;
{
  DisplayCB *t;

  t = (DisplayCB *)MPGet(wd->mp, sizeof(DisplayCB));
  memcpy(&(t->dcb), dcb, sizeof(t->dcb));
  t->closure = dc;

  GListAddHead(wd->cblist, t);

  return;
}

/*
 * WWWClearDisplayCallbacks
 */
void
WWWClearDisplayCallbacks(wd)
WWWDisplay wd;
{
  while (GListPop(wd->cblist) != NULL)
      ;
  return;
}
