/*
 * render.c
 *
 * Copyright (c) 1995-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "WWWP.h"

typedef struct
{
  WWWRenderCallbacks cb;
  void        *closure;
} RenderCB;

struct WWWRenderP
{
  MemPool     mp;                     /* memory pool */
  WWWRenderHooks rh;                  /* renderer hooks */
  void        *rc;                    /* renderer context */
  GList       cblist;
  WWWContext  wc;
  WWWDisplay  wd;
  WWWData     wp;
};

/*
 * WWWDestroyRender
 */
void
WWWDestroyRender(wn)
WWWRender wn;
{
  (wn->rh.destroy)(wn->rc);
  MPDestroy(wn->mp);
  return;
}

/*
 * WWWAddRender
 */
void
WWWAddRender(wn)
WWWRender wn;
{
  (wn->rh.add)(wn->rc);
  return;
}

/*
 * WWWEndRender
 */
void
WWWEndRender(wn)
WWWRender wn;
{
  (wn->rh.end)(wn->rc);
  return;
}

/*
 * WWWQueryRender
 */
byte *
WWWQueryRender(wn, key, size)
WWWRender wn;
byte *key;
size_t *size;
{
  if (wn->rh.query != NULL) return((wn->rh.query)(wn->rc, key, size));
  return(NULL);
}

/*
 * WWWSearchRender
 */
int
WWWSearchRender(wn, data, len)
WWWRender wn;
byte *data;
size_t len;
{
  if (wn->rh.search != NULL) return((wn->rh.search)(wn->rc, data, len));
  return(-1);
}

/*
 * WWWCancelRender
 */
void
WWWCancelRender(wn)
WWWRender wn;
{
  (wn->rh.cancel)(wn->rc);
  return;
}

/*
 * WWWCreateRender
 */
WWWRender
WWWCreateRender(wc, wd, wp, rh)
WWWContext wc;
WWWDisplay wd;
WWWData wp;
WWWRenderHooks *rh;
{
  WWWRender wn;
  MemPool mp;

  myassert(rh->init != NULL, "Callback is NULL.");
  myassert(rh->add != NULL, "Callback is NULL.");
  myassert(rh->end != NULL, "Callback is NULL.");
  myassert(rh->destroy != NULL, "Callback is NULL.");
  myassert(rh->cancel != NULL, "Callback is NULL.");

  mp = MPCreate();
  wn = (WWWRender)MPCGet(mp, sizeof(struct WWWRenderP));
  wn->mp = mp;
  wn->wc = wc;
  wn->wd = wd;
  wn->wp = wp;
  wn->cblist = GListCreateX(mp);

  memcpy(&(wn->rh), rh, sizeof(WWWRenderHooks));

  if ((wn->rc = (rh->init)(wn, rh->class_context)) == NULL)
  {
    MPDestroy(wn->mp);
    return(NULL);
  }

  return(wn);
}

void
WWWDoRenderAction(wn, ai)
WWWRender wn;
WWWRenderActionInfo *ai;
{
  RenderCB *t;

  for (t = (RenderCB *)GListGetHead(wn->cblist); t != NULL;
       t = (RenderCB *)GListGetNext(wn->cblist))
  {
    if (t->cb.action != NULL)
    {
      if (!(t->cb.action)(t->closure, ai)) return;
    }
  }
  return;
}

WWWDisplay
WWWGetRenderDisplay(wn)
WWWRender wn;
{
  return(wn->wd);
}

WWWData
WWWGetRenderData(wn)
WWWRender wn;
{
  return(wn->wp);
}

WWWContext
WWWGetRenderContext(wn)
WWWRender wn;
{
  return(wn->wc);
}

void
WWWSetRenderCallbacks(wn, rcb, rc)
WWWRender wn;
WWWRenderCallbacks *rcb;
void *rc;
{
  RenderCB *t;

  t = (RenderCB *)MPGet(wn->mp, sizeof(RenderCB));
  memcpy(&(t->cb), rcb, sizeof(t->cb));
  t->closure = rc;

  GListAddHead(wn->cblist, t);

  return;
}

void
WWWSendRenderMessage(wn, message)
WWWRender wn;
char *message;
{
  RenderCB *t;

  for (t = (RenderCB *)GListGetHead(wn->cblist); t != NULL;
       t = (RenderCB *)GListGetNext(wn->cblist))
  {
    if (t->cb.message != NULL)
    {
      if (!(t->cb.message)(t->closure, message)) return;
    }
  }
  return;
}
